@@ -16,10 +16,8 @@
  *    -stibor-
  */
 #include "config.h"
-#include "ylib.h"
 #include "wmapp.h"
 #include "applet.h"
-#include "ypointer.h"
 #include "acpustatus.h"
 
 #if IWM_STATES
@@ -32,15 +30,15 @@
 #if __linux__
 #include <sys/sysinfo.h>
 #else
-
+#if __FreeBSD__
+#include <sys/resource.h>
+#endif
 #if __sun__
 #include <sys/loadavg.h>
 #endif
 
-#ifdef HAVE_SYS_PARAM_H
 #include <sys/param.h>
-#endif
-#ifdef HAVE_SYS_SYSCTL_H
+#if HAVE_SYS_SYSCTL_H || HAVE_SYSCTLBYNAME || __FreeBSD__
 #include <sys/sysctl.h>
 #endif
 #ifdef HAVE_UVM_UVM_PARAM_H
@@ -66,15 +64,100 @@
 
 extern ref<YPixmap> taskbackPixmap;
 
-ref<YFont> CPUStatus::tempFont;
+#if __linux__
+class YStat {
+public:
+    int count() {
+        int cpus = 0;
+        if (have()) {
+            char* p = data;
+            while (p && (p = strstr(p, "\ncpu")) != nullptr) {
+                cpus += ASCII::isDigit(p[4]);
+                p = strchr(p + 4, '\n');
+            }
+        }
+        return cpus;
+    }
+    void recycle() {
+        if (data) {
+            *data = '\0';
+        }
+    }
+    char* cpustatus(int cpuid) {
+        char* status = nullptr;
+        if (have()) {
+            if (cpuid == -1) {
+                char* p = strstr(data, "cpu ");
+                if (p == data || p[-1] == '\n') {
+                    status = p + 4;
+                }
+            } else {
+                char buf[32];
+                snprintf(buf, sizeof buf, "\ncpu%d ", cpuid);
+                char* p = strstr(data, buf);
+                if (p) {
+                    status = p + strlen(buf);
+                }
+            }
+        }
+        return status;
+    }
+    YStat() : fildes(-1), data(nullptr) { }
+    ~YStat() {
+        if (fildes > 2) close(fildes);
+        if (data) delete[] data;
+    }
+private:
+    int fildes;
+    char* data;
+    const size_t size = 8192;
+    bool have() { return nonempty(data) || read(); }
+    bool read() {
+        const char path[] = "/proc/stat";
+        if (fildes >= 0 && lseek(fildes, 0, SEEK_SET) < 0) {
+            if (ONCE)
+                fail("seek %s", path);
+            close(fildes);
+            fildes = -1;
+        }
+        if (fildes < 0 && (fildes = open(path, O_RDONLY | O_CLOEXEC)) < 0) {
+            if (ONCE)
+                fail("open %s", path);
+            return false;
+        }
+        if (data == nullptr) {
+            data = new char[size];
+            *data = '\0';
+        }
+        ssize_t len = ::read(fildes, data, size - 1);
+        if (len <= 0) {
+            if (ONCE)
+                fail("read %s", path);
+            close(fildes);
+            fildes = -1;
+            delete[] data;
+            data = nullptr;
+            return false;
+        } else {
+            data[len] = '\0';
+        }
+        return true;
+    }
+};
+#endif
 
-CPUStatus::CPUStatus(YWindow *aParent, CPUStatusHandler *aHandler, int cpuid) :
-    IApplet(this, aParent),
+YFont CPUStatus::tempFont;
+YTemp* CPUStatus::fTemp;
+
+CPUStatus::CPUStatus(YWindow* parent, CPUStatusHandler* handler,
+                     int cpuid, YStat* stat) :
+    IApplet(this, parent),
     fCpuID(cpuid),
     statusUpdateCount(0),
     unchanged(taskBarCPUSamples),
     cpu(taskBarCPUSamples, IWM_STATES),
-    fHandler(aHandler),
+    fHandler(handler),
+    fStat(stat),
     fTempColor(&clrCpuTemp)
 {
     cpu.clear();
@@ -91,12 +174,9 @@ CPUStatus::CPUStatus(YWindow *aParent, CPUStatusHandler *aHandler, int cpuid) :
     color[IWM_IDLE] = &clrCpuIdle;
     color[IWM_STEAL] = &clrCpuSteal;
 
-    fUpdateTimer->setTimer(taskBarCPUDelay, this, true);
-
     setSize(taskBarCPUSamples, taskBarGraphHeight);
     getStatus();
     updateStatus();
-    updateToolTip();
     char buf[99];
     snprintf(buf, 99, "CPU%d", cpuid);
     setTitle(buf);
@@ -110,10 +190,13 @@ bool CPUStatus::picture() {
 
     Graphics G(getPixmap(), width(), height(), depth());
 
+#if __linux__
     if (cpustatusShowAcpiTempInGraph) {
         change = true;
         statusUpdateCount = taskBarCPUSamples;
     }
+#endif
+
     if (change) {
         unchanged = 0;
         fill(G);
@@ -123,9 +206,13 @@ bool CPUStatus::picture() {
         draw(G);
         change = true;
     }
+
+#if __linux__
     if (cpustatusShowAcpiTempInGraph) {
         temperature(G);
     }
+#endif
+
     return change;
 }
 
@@ -252,25 +339,192 @@ void CPUStatus::draw(Graphics& g) {
     }
 }
 
+#if __linux__
+class YTemp : private YTimerListener {
+public:
+    YTemp() :
+        // TRANSLATORS: Please translate the string "C" into "Celsius Temperature" in your language.
+        // TRANSLATORS: Please make sure the translated string could be shown in your non-utf8 locale.
+        centigrade(_("°C"))
+    {
+        read_types();
+    }
+    bool read_temp(char* tempbuf, int buflen, bool best) {
+        tempbuf[0] = '\0';
+        int done = 0;
+        for (int k = 0; k < thermal.getCount() && done + 6 < buflen; ++k) {
+            int temp = read_data(k);
+            if (temp) {
+                int cent = temp / 1000;
+                int hund = temp % 1000 / 100;
+                if (best && ! hund) {
+                    snprintf(tempbuf + done, buflen - done, "%d ", cent);
+                } else {
+                    snprintf(tempbuf + done, buflen - done, "%d.%d ", cent, hund);
+                }
+                done += strlen(tempbuf + done);
+                if (best) {
+                    break;
+                }
+            }
+        }
+        if (done && done + int(strlen(centigrade)) + 1 < buflen) {
+            strlcpy(tempbuf + done - best, centigrade, buflen - done);
+        }
+        return done > 0;
+    }
+private:
+    bool handleTimer(YTimer *timer) {
+        for (int k = 0; k < thermal.getCount(); ++k) {
+            thermal[k].temp = 0;
+        }
+        return false;
+    }
+    bool read_data(const char* name) {
+        buf[0] = '\0';
+        char path[64];
+        strlcpy(path, "/sys/class/thermal/", sizeof path);
+        strlcat(path, name, sizeof path);
+        if (strlcat(path, "/temp", sizeof path) < sizeof path) {
+            int fd = open(path, O_RDONLY);
+            if (fd > 2) {
+                if (read(fd, buf, sizeof buf) < 1) {
+                    buf[0] = '\0';
+                }
+                close(fd);
+            }
+        }
+        return buf[0] != '\0';
+    }
+    int read_data(int index) {
+        int data = thermal[index].temp;
+        if (data == 0) {
+            if (read_data(thermal[index].name)) {
+                thermal[index].temp = data = atoi(buf);
+                if (data && timer.isRunning() == false) {
+                    timer.setTimer(max(1000, taskBarCPUDelay) - 1, this, true);
+                }
+            }
+        }
+        return data;
+    }
+    bool read_type(const char* name, int dirfd) {
+        buf[0] = '\0';
+        if (strlcpy(buf, name, sizeof buf) + 5 < sizeof buf) {
+            strlcat(buf, "/type", sizeof buf);
+            int fd = openat(dirfd, buf, O_RDONLY);
+            if (fd > 2) {
+                memset(buf, 0, sizeof buf);
+                if (read(fd, buf, sizeof buf - 1) < 1) {
+                    buf[0] = '\0';
+                }
+                close(fd);
+                char* newl = strchr(buf, '\n');
+                if (newl) {
+                    *newl = '\0';
+                }
+            }
+        }
+        return buf[0] != '\0';
+    }
+    int prioritize(const char* name) {
+        if (strncmp(buf, "x86_pkg_temp", 12) == 0) {
+            if (read_data(name)) {
+                return 4;
+            }
+        }
+        else if (strncmp(buf, "pch_", 4) == 0) {
+            if (read_data(name)) {
+                return 3;
+            }
+        }
+        else if (strcmp(buf, "acpitz") == 0) {
+            if (read_data(name)) {
+                return 2;
+            }
+        }
+        return 0;
+    }
+    void read_types() {
+        for (cdir dir("/sys/class/thermal"); dir.next(); ) {
+            const char* name = dir.entry();
+            if (strncmp(name, "thermal", 7) == 0 && int(strlen(name)) < 24) {
+                if (read_type(name, dir.descriptor())) {
+                    int prio = prioritize(name);
+                    if (prio) {
+                        int k = 0;
+                        while (k < thermal.getCount()
+                            && prio <= thermal[k].prio) {
+                            ++k;
+                        }
+                        if (k < 6) {
+                            thermal.insert(k, priotemp(prio, name));
+                        }
+                        if (thermal.getCount() > 6) {
+                            thermal.shrink(6);
+                        }
+                    }
+                }
+            }
+        }
+    }
+    YTimer timer;
+    char buf[24];
+    const char* centigrade;
+    struct priotemp {
+        int prio; int temp; char name[24];
+        priotemp(int p, const char* n) : prio(p), temp(0) {
+            strlcpy(name, n, sizeof name);
+        }
+    };
+    YArray<priotemp> thermal;
+};
+#endif
+
+void CPUStatus::freeTemp() {
+#if __linux__
+    if (fTemp) {
+        delete fTemp;
+        fTemp = nullptr;
+    }
+#endif
+}
+
 void CPUStatus::temperature(Graphics& g) {
+#if __linux__
     if (cpustatusShowAcpiTempInGraph) {
         char temp[10];
-        getAcpiTemp(temp, sizeof(temp));
+        getAcpiTemp(temp, sizeof(temp), true);
         g.setColor(fTempColor);
         if (tempFont == null)
-            tempFont = YFont::getFont(XFA(tempFontName));
-        g.setFont(tempFont);
-        int h = height();
-        int y = (h - 1 - tempFont->height()) / 2 + tempFont->ascent();
-        int w = tempFont->textWidth(temp);
-        int x = max(0, (int(g.rwidth()) - w) / 2);
-        g.drawChars(temp, 0, 3, x, y);
+            tempFont = tempFontName;
+        if (tempFont) {
+            g.setFont(tempFont);
+            int h = height();
+            int y = (h - 1 - tempFont->height()) / 2 + tempFont->ascent();
+            int w = tempFont->textWidth(temp);
+            if (w > int(width())) {
+                int s = strspn(temp, "0123456789.");
+                if (s && temp[s]) {
+                    temp[s] = '\0';
+                    w = tempFont->textWidth(temp);
+                }
+                while (w > int(width()) && strchr(temp, '.') && 0 < s) {
+                    temp[--s] = '\0';
+                    if (temp[s-1] == '.') {
+                        temp[--s] = '\0';
+                    }
+                    w = tempFont->textWidth(temp);
+                }
+            }
+            int x = max(0, (int(g.rwidth()) - w) / 2);
+            g.drawString(x, y, temp);
+        }
     }
+#endif
 }
 
 bool CPUStatus::handleTimer(YTimer *t) {
-    if (t != fUpdateTimer)
-        return false;
     if (toolTipVisible())
         updateToolTip();
     updateStatus();
@@ -278,18 +532,20 @@ bool CPUStatus::handleTimer(YTimer *t) {
 }
 
 void CPUStatus::updateToolTip() {
+    char buf[256];
     char cpuid[16] = "";
     if (fCpuID >= 0)
         snprintf(cpuid, sizeof(cpuid), "%d", fCpuID);
+
 #ifdef __linux__
-    char fmt[255] = "";
+    memset(buf, 0, sizeof buf);
 #define ___checkspace if (more<0 || rest-more<=0) return; pos+=more; rest-=more;
     struct sysinfo sys;
 
-    if (0==sysinfo(&sys))
+    if (0 == sysinfo(&sys))
     {
-        char *pos=fmt;
-        int rest=sizeof(fmt);
+        char* pos = buf;
+        int rest = sizeof(buf);
         float l1 = float(sys.loads[0]) / 65536.0,
               l5 = float(sys.loads[1]) / 65536.0,
               l15 = float(sys.loads[2]) / 65536.0;
@@ -308,18 +564,22 @@ void CPUStatus::updateToolTip() {
                     GBnorm(sys.totalswap), GBnorm(sys.freeswap));
             ___checkspace;
         }
+
+#if __linux__
         if (cpustatusShowAcpiTemp) {
-            char *posEx=pos;
+            char *posEx = pos;
             more=snprintf(pos, rest, _("\nACPI Temp: "));
             ___checkspace;
-            more=getAcpiTemp(pos, rest);
+            more = getAcpiTemp(pos, rest, false);
             if (more)
             {
               ___checkspace;
             }
             else
-               pos=posEx;
+               pos = posEx;
         }
+#endif
+
         if (cpustatusShowCpuFreq) {
             float maxf = getCpuFreq(0), minf = maxf;
             int cpus;
@@ -338,10 +598,8 @@ void CPUStatus::updateToolTip() {
                         form, maxf / 1e6, minf / 1e6, perc + 4);
             }
         }
-        setToolTip(mstring(fmt));
     }
 #else
-    char buf[99];
     double loadavg[3];
     if (getloadavg(loadavg, 3) < 0)
         return;
@@ -371,9 +629,31 @@ void CPUStatus::updateToolTip() {
         snprintf(buf + strlen(buf), sizeof buf - strlen(buf),
                  _("\nCPU Freq: %.3fGHz"), freq * 1e-9);
     }
+#endif
+
+    for (char* str = buf; *str; ) {
+        char* nln = strchr(str, '\n');
+        int len = nln ? int(nln - str) : int(strlen(str));
+        char* tab = static_cast<char *>(memchr(str, '\t', len));
+        if (tab) {
+        } else {
+            char* col = static_cast<char *>(memchr(str, ':', len));
+            if (col) {
+                if (col[1] == ' ') {
+                    col[1] = '\t';
+                } else {
+                    size_t size = strlen(col);
+                    if (col - buf + size + 2 < sizeof buf) {
+                        memmove(col + 2, col + 1, size);
+                        col[1] = '\t';
+                    }
+                }
+            }
+        }
+        str = nln ? nln + 1 : str + len;
+    }
 
     setToolTip(buf);
-#endif
 }
 
 void CPUStatus::handleClick(const XButtonEvent &up, int count) {
@@ -394,67 +674,22 @@ void CPUStatus::updateStatus() {
     repaint();
 }
 
-int CPUStatus::getAcpiTemp(char *tempbuf, int buflen) {
-    int retbuflen = 0;
-    char namebuf[300];
-    char buf[64];
-
-    memset(tempbuf, 0, buflen);
+int CPUStatus::getAcpiTemp(char *tempbuf, int buflen, bool best) {
+    tempbuf[0] = '\0';
 
 #if __linux__
-    cdir dir;
-    if (dir.open("/sys/class/thermal")) {
-        while (dir.next()) {
-            if (strncmp(dir.entry(), "thermal", 7))
-                continue;
-            int len;
-
-            snprintf(namebuf, sizeof namebuf,
-                    "/sys/class/thermal/%s/temp", dir.entry());
-            len = read_file(namebuf, buf, sizeof(buf));
-            if (len > 4) {
-                int seglen = len - 4;
-                if (retbuflen + seglen + 4 >= buflen) {
-                    break;
-                }
-                strncat(tempbuf + retbuflen, buf, seglen);
-                retbuflen += seglen;
-                tempbuf[retbuflen++] = '.';
-                tempbuf[retbuflen++] = buf[seglen];
-                tempbuf[retbuflen++] = ' ';
-                tempbuf[retbuflen] = '\0';
-            }
-        }
-        if (1 < retbuflen && retbuflen + 1 < buflen) {
-            // TRANSLATORS: Please translate the string "C" into "Celsius Temperature" in your language.
-            // TRANSLATORS: Please make sure the translated string could be shown in your non-utf8 locale.
-            static const char *T = _("°C");
-            int i = -1;
-            while (T[++i]) tempbuf[retbuflen++] = T[i];
-            tempbuf[retbuflen] = '\0';
-        }
+    if (fTemp == nullptr) {
+        fTemp = new YTemp;
     }
-    else if (dir.open("/proc/acpi/thermal_zone")) {
-        while (dir.next()) {
-            int len, seglen = 7;
-            snprintf(namebuf, sizeof namebuf,
-                    "/proc/acpi/thermal_zone/%s/temperature", dir.entry());
-            len = read_file(namebuf, buf, sizeof(buf));
-            if (len > seglen) {
-                if (retbuflen + seglen >= buflen) {
-                    break;
-                }
-                retbuflen += seglen;
-                strncat(tempbuf, buf + len - seglen, seglen);
-            }
-        }
+    if (fTemp) {
+        fTemp->read_temp(tempbuf, buflen, best);
     }
 #endif
 
-    return retbuflen;
+    return int(strlen(tempbuf));
 }
 
-float CPUStatus::getCpuFreq(unsigned int cpu) {
+float CPUStatus::getCpuFreq(int cpu) {
 
 #if __linux__
     char buf[16], namebuf[100];
@@ -462,10 +697,10 @@ float CPUStatus::getCpuFreq(unsigned int cpu) {
     for (unsigned i = 0; i < ACOUNT(categories); ++i)
     {
         float cpufreq = 0;
-        snprintf(namebuf, sizeof namebuf,
+        snprintf(BUFNSIZE(namebuf),
                 "/sys/devices/system/cpu/cpu%d/cpufreq/%s_cur_freq",
                 cpu, categories[i]);
-        if (read_file(namebuf, buf, sizeof(buf)) > 0) {
+        if (filereader(namebuf).read_all(BUFNSIZE(buf)) > 0) {
             sscanf(buf, "%f", &cpufreq);
             return cpufreq;
         }
@@ -488,30 +723,12 @@ float CPUStatus::getCpuFreq(unsigned int cpu) {
 }
 
 
-void CPUStatus::getStatusPlatform() {
+void CPUStatus::getStatusLinux() {
 #ifdef __linux__
-    char *p = nullptr, buf[4096], *end = nullptr;
+    char *p = fStat->cpustatus(fCpuID);
     unsigned long long cur[IWM_STATES];
     int s;
 
-    fileptr fd(fopen("/proc/stat", "r"));
-    if (fd == nullptr)
-        return;
-
-    while (fgets(buf, sizeof buf, fd) && 0 == strncmp(buf, "cpu", 3)) {
-        if (fCpuID == -1) {
-            if (ASCII::isSpaceOrTab(buf[3]))
-                p = buf + 4;
-            break;
-        }
-        if (ASCII::isDigit(buf[3]) && strtol(buf + 3, &end, 10) == fCpuID) {
-            if (end > buf && ASCII::isSpaceOrTab(*end)) {
-                p = 1 + end;
-                break;
-            }
-        }
-    }
-    fd.close();
     if (p == nullptr)
         return;
 
@@ -541,9 +758,12 @@ void CPUStatus::getStatusPlatform() {
         cpu[taskBarCPUSamples - 1][i] = cur[i] - last_cpu[i];
         last_cpu[i] = cur[i];
     }
+#endif
+}
 
-    return;
-
+void CPUStatus::getStatusPlatform() {
+#if __linux__
+    getStatusLinux();
 #elif __OpenBSD__ || __NetBSD__ || __FreeBSD__
 
 #if defined __NetBSD__
@@ -620,26 +840,35 @@ CPUStatusControl::CPUStatusControl(YSMListener *smActionListener,
     smActionListener(smActionListener),
     iapp(iapp),
     aParent(aParent),
+    fStat(nullptr),
     fMenuCPU(-1),
+    fSamples(taskBarCPUSamples),
     fPid(0)
 {
     GetCPUStatus(cpuCombine);
+    fUpdateTimer.setTimer(taskBarCPUDelay, this, true);
+}
+
+CPUStatusControl::~CPUStatusControl() {
+    CPUStatus::freeFont();
+#if __linux__
+    CPUStatus::freeTemp();
+    delete fStat;
+#endif
 }
 
 void CPUStatusControl::GetCPUStatus(bool combine) {
     int count = 0;
 #if __linux__
-    if (combine == false) {
-        fileptr fd("/proc/stat", "r");
-        if (fd) {
-            char buf[128];
-            while (fgets(buf, sizeof buf, fd) && 0 == memcmp(buf, "cpu", 3))
-                count += ASCII::isDigit(buf[3]);
-        }
+    if (fStat == nullptr) {
+        fStat = new YStat();
+    }
+    if (combine == false && fStat) {
+        count = fStat->count();
     }
 #endif
     do {
-        fCPUStatus += new CPUStatus(aParent, this, --count);
+        fCPUStatus += new CPUStatus(aParent, this, --count, fStat);
     } while (0 < count);
 }
 
@@ -648,6 +877,27 @@ void CPUStatusControl::runCommandOnce(const char *resource, const char *cmdline)
     smActionListener->runCommandOnce(resource, cmdline, &fPid);
 }
 
+bool CPUStatusControl::handleTimer(YTimer* timer) {
+    if (timer == &fUpdateTimer) {
+        if (fSamples != taskBarCPUSamples) {
+            fSamples = taskBarCPUSamples;
+            int cpus = fCPUStatus.getCount();
+            fCPUStatus.clear();
+            GetCPUStatus(cpus < 2);
+            iapp->relayout();
+        }
+#if __linux__
+        if (fStat) {
+            fStat->recycle();
+        }
+#endif
+        for (CPUStatus* status : fCPUStatus) {
+            status->handleTimer(timer);
+        }
+    }
+    return true;
+}
+
 void CPUStatusControl::handleClick(const XButtonEvent &up, int cpuid) {
     if (up.button == Button3) {
         fMenu = new YMenu();
