@@ -4,32 +4,22 @@
  * Copyright (C) 1997-2002 Marko Macek
  */
 #include "config.h"
-
-#include "yfull.h"
-#include "ykey.h"
 #include "ymenu.h"
-#include "yaction.h"
 #include "ymenuitem.h"
-#include "yrect.h"
-
 #include "yxapp.h"
 #include "prefs.h"
 #include "yprefs.h"
-
 #include "ascii.h"
-
-#include "yicon.h"
-
 #include <string.h>
 
-YColorName menuBg(&clrNormalMenu);
-YColorName menuItemFg(&clrNormalMenuItemText);
-YColorName activeMenuItemBg(&clrActiveMenuItem);
-YColorName activeMenuItemFg(&clrActiveMenuItemText);
-YColorName disabledMenuItemFg(&clrDisabledMenuItemText);
-YColorName disabledMenuItemSt(&clrDisabledMenuItemShadow);
+static YColorName menuBg(&clrNormalMenu);
+static YColorName menuItemFg(&clrNormalMenuItemText);
+static YColorName activeMenuItemBg(&clrActiveMenuItem);
+static YColorName activeMenuItemFg(&clrActiveMenuItemText);
+static YColorName disabledMenuItemFg(&clrDisabledMenuItemText);
+static YColorName disabledMenuItemSt(&clrDisabledMenuItemShadow);
 
-ref<YFont> menuFont;
+YFont menuFont;
 
 ref<YPixmap> menusepPixmap;
 ref<YPixmap> menuselPixmap;
@@ -43,21 +33,20 @@ int YMenu::fAutoScrollDeltaX = 0;
 int YMenu::fAutoScrollDeltaY = 0;
 int YMenu::fAutoScrollMouseX = -1;
 int YMenu::fAutoScrollMouseY = -1;
+int YMenu::fMenuObjectCount;
 YMenu *YMenu::fPointedMenu = nullptr;
 
 void YMenu::setActionListener(YActionListener *actionListener) {
     fActionListener = actionListener;
 }
 
-void YMenu::finishPopup(YMenuItem *item, YAction action,
-                        unsigned int modifiers)
-{
+void YMenu::finishPopup(YMenuItem *item, YAction action, unsigned modifiers) {
     YActionListener *cmd = fActionListener;
 
     YPopupWindow::finishPopup();
 
-    if (item)
-        item->actionPerformed(cmd, action, modifiers);
+    if (cmd)
+        cmd->actionPerformed(action, modifiers);
 }
 
 lazy<YTimer> YMenu::fMenuTimer;
@@ -68,10 +57,8 @@ YMenu::YMenu(YWindow *parent):
     fGradient(null),
     fMenusel(null)
 {
-    if (menuFont == null)
-        menuFont = YFont::getFont(XFA(menuFontName));
-
     paintedItem = selectedItem = -1;
+    paramPos = namePos = 0;
     submenuItem = -1;
     fPopup = nullptr;
     fActionListener = nullptr;
@@ -84,6 +71,10 @@ YMenu::YMenu(YWindow *parent):
     fTimerY = 0;
     fTimerSubmenuItem = -1;
     addStyle(wsNoExpose);
+
+    if (menuFont == null)
+        menuFont = menuFontName;
+    ++fMenuObjectCount;
 }
 
 void YMenu::raise() {
@@ -105,6 +96,8 @@ YMenu::~YMenu() {
     fGradient = null;
     fMenusel = null;
     removeAll();
+    if (--fMenuObjectCount == 0)
+        menuFont = null;
 }
 
 void YMenu::activatePopup(int flags) {
@@ -158,7 +151,7 @@ int YMenu::onCascadeButton(int selItem, int x, int /*y*/, bool /*checkPopup*/) {
             return 0;
 #endif
 
-        unsigned fontHeight = menuFont->height() + 1;
+        unsigned fontHeight = (menuFont ? menuFont->height() : 0) + 1;
         unsigned h = fontHeight;
 
         if (getItem(selItem)->getIcon() != null &&
@@ -219,15 +212,15 @@ void YMenu::activateSubMenu(int item, bool byMouse) {
             getOffsets(l, t, r, b);
             findItemPos(item, xp, yp, ih);
             YRect rect(x(), y(), width(), height());
-            if (sub->getActionListener() == nullptr)
-                sub->setActionListener(getActionListener());
+            if (sub->fActionListener == nullptr)
+                sub->setActionListener(fActionListener);
             sub->popup(nullptr, this, nullptr,
                        x() + int(width()) - r, y() + yp - t,
                        int(width()) - r - l, -1,
                        getXiScreen(),
-                       YPopupWindow::pfCanFlipHorizontal |
-                       (popupFlags() & YPopupWindow::pfFlipHorizontal) |
-                       (byMouse ? (unsigned int)YPopupWindow::pfButtonDown : 0U));
+                       pfCanFlipHorizontal |
+                       (popupFlags() & pfFlipHorizontal) |
+                       (byMouse ? pfButtonDown : 0));
             fPopup = sub;
             submenuItem = item;
         }
@@ -327,7 +320,7 @@ bool YMenu::handleKey(const XKeyEvent &key) {
     if (key.type == KeyPress) {
         if ((m & ~ShiftMask) == 0) {
             if (k == XK_Escape) {
-                cancelPopup();
+                YPopupWindow::finishPopup();
             } else if (k == XK_Left || k == XK_KP_Left) {
                 if (prevPopup())
                     cancelPopup();
@@ -392,7 +385,8 @@ void YMenu::handleButton(const XButtonEvent &button) {
         if (button.type == ButtonPress && itemCount() > 0) {
             if (inrange(button.x_root, x(), x() + int(width()) - 1)) {
                 const int itemHeight = height() / itemCount();
-                const int stepSize = max(menuFont->height(), itemHeight);
+                const int fontHeight = menuFont ? menuFont->height() : 0;
+                const int stepSize = max(fontHeight, itemHeight);
                 hideSubmenu();
                 setPosition(x(), clamp(y() - (int)(button.state & ShiftMask ?
                                                    3 * stepSize : stepSize),
@@ -409,7 +403,8 @@ void YMenu::handleButton(const XButtonEvent &button) {
         if (button.type == ButtonPress && itemCount() > 0) {
             if (inrange(button.x_root, x(), x() + int(width()) - 1)) {
                 const int itemHeight = height() / itemCount();
-                const int stepSize = max(menuFont->height(), itemHeight);
+                const int fontHeight = menuFont ? menuFont->height() : 0;
+                const int stepSize = max(fontHeight, itemHeight);
                 hideSubmenu();
                 setPosition(x(), clamp(y() + (int)(button.state & ShiftMask ?
                                                    3 * stepSize : stepSize),
@@ -662,23 +657,54 @@ YMenuItem *YMenu::addSubmenu(const mstring &name, int hotCharPos, YMenu *submenu
     return add(new YMenuItem(name, hotCharPos, null, actionNull, submenu));
 }
 
-YMenuItem * YMenu::addSeparator() {
-    return add(new YMenuItem());
+YMenuItem *YMenu::addSubmenu(const mstring &name, int hotCharPos,
+                             YMenu *submenu, YMenuItem* after)
+{
+    YMenuItem* item = new YMenuItem(name, hotCharPos, null, actionNull, submenu);
+    return add(item, after);
+}
+
+void YMenu::addSeparator() {
+    int n = itemCount();
+    if (n > 0 && !fItems[n - 1]->isSeparator()) {
+        void(add(new YMenuItem()));
+    }
 }
 
 YMenuItem *YMenu::addLabel(const mstring &name) {
     return add(new YMenuItem(name));
 }
 
+bool YMenu::lastIsSeparator() const {
+    int n = itemCount();
+    return n > 0 && fItems[n - 1]->isSeparator();
+}
+
+YMenuItem* YMenu::lastItem() const {
+    int n = itemCount();
+    return n > 0 ? fItems[n - 1] : nullptr;
+}
+
+bool YMenu::removeSubmenu(YMenu* menu) {
+    bool done = false;
+    YMenuItem* item = findSubmenu(menu);
+    if (item) {
+        hideSubmenu();
+        int index = find(fItems, item);
+        if (index >= 0) {
+            fItems[index]->setSubmenu(nullptr);
+            fItems.remove(index);
+        }
+        done = true;
+    }
+    return done;
+}
+
 void YMenu::removeAll() {
     hideSubmenu();
-    const int n = itemCount();
-    for (int i = 0; i < n; ++i) {
-        YMenu* sub = fItems[i]->getSubmenu();
-        if (sub && sub->isShared()) {
-            YMenuItem* nil(nullptr);
-            swap(nil, fItems[i]);
-        }
+    for (YMenuItem* item : fItems) {
+        if (item->getSubmenu() && item->getSubmenu()->isShared())
+            item->setSubmenu(nullptr);
     }
     fItems.clear();
     // paintedItem = selectedItem = -1;
@@ -689,90 +715,117 @@ YMenuItem * YMenu::add(YMenuItem *item) {
     return item;
 }
 
+YMenuItem * YMenu::add(YMenuItem* item, YMenuItem* after) {
+    if (item) {
+        int i = find(fItems, after);
+        if (i >= 0)
+            fItems.insert(i + 1, item);
+        else
+            fItems.append(item);
+    }
+    return item;
+}
+
 YMenuItem * YMenu::add(YMenuItem *item, const char *icons) {
-    //YIcon *icon = 0;
-    ref<YIcon> icon = YIcon::getIcon(icons);
-    if (icon->isCached() && item) item->setIcon(icon);
-    if (item) fItems.append(item);
+    if (item) {
+        ref<YIcon> icon(YIcon::getIcon(icons));
+        if (icon != null && icon->isCached()) {
+            item->setIcon(icon);
+        }
+        fItems.append(item);
+    }
     return item;
 }
 
 YMenuItem * YMenu::addSorted(YMenuItem *item, bool duplicates, bool ignoreCase) {
-    for (int i = 0; i < itemCount(); i++) {
-        if (item->getName() == null || fItems[i]->getName() == null)
-            continue;
-
-        int cmp = item->getName().collate(fItems[i]->getName(), ignoreCase);
-        if (cmp > 0)
-            continue;
-        else if (cmp != 0 || duplicates) {
-            fItems.insert(i, item);
-            return item;
-        } else {
-            return nullptr;
+    if (item) {
+        if (item->haveName()) {
+            mstring& name = item->getName();
+            for (int i = 0; i < itemCount(); i++) {
+                if (fItems[i]->haveName()) {
+                    int cmp = name.collate(fItems[i]->getName(), ignoreCase);
+                    if (cmp > 0)
+                        continue;
+                    else if (cmp != 0 || duplicates) {
+                        fItems.insert(i, item);
+                        return item;
+                    } else {
+                        return nullptr;
+                    }
+                }
+            }
         }
+        fItems.append(item);
     }
-    if (item) fItems.append(item);
     return item;
 }
 
+bool YMenu::haveCommand(YAction action) {
+    return findAction(action) != nullptr;
+}
+
 YMenuItem *YMenu::findAction(YAction action) {
-    for (int i = 0; i < itemCount(); i++)
-        if (action == getItem(i)->getAction()) return getItem(i);
+    for (YMenuItem* item : fItems)
+        if (action == item->getAction())
+            return item;
     return nullptr;
 }
 
 YMenuItem *YMenu::findSubmenu(const YMenu *sub) {
-    for (int i = 0; i < itemCount(); i++)
-        if (sub == getItem(i)->getSubmenu()) return getItem(i);
+    for (YMenuItem* item : fItems)
+        if (sub == item->getSubmenu())
+            return item;
     return nullptr;
 }
 
 YMenuItem *YMenu::findName(const mstring &name, const int first) {
     if (name != null)
-        for (int i = first; i < itemCount(); i++) {
-            mstring iname = getItem(i)->getName();
-            if (iname != null && iname.equals(name))
-                return getItem(i);
-    }
+        for (YMenuItem* item : fItems)
+            if (item->haveName(name))
+                return item;
 
     return nullptr;
 }
 
-int YMenu::findFirstLetRef(char firstLet, const int first, const int ignCase) {
-    if (ignCase)
-        firstLet = ASCII::toUpper(firstLet);
+int YMenu::findFirstLetRef(char firstLetter, int first, bool ignoreCase) {
+    if (ignoreCase)
+        firstLetter = ASCII::toUpper(firstLetter);
     for (int i = first; i < itemCount(); i++) {
-        YMenuItem *temp = getItem(i);
-        mstring iLetterRef = temp->getName();
-        if (iLetterRef != null) {
-            int iLetter = iLetterRef.charAt(0);
-            if (ignCase)
+        YMenuItem *item = getItem(i);
+        if (item->haveName()) {
+            int iLetter = item->getName().charAt(0);
+            if (ignoreCase)
                 iLetter = ASCII::toUpper(iLetter);
-            if (iLetter == firstLet)
+            if (iLetter == firstLetter)
                 return i;
         }
     }
     return -1;
 }
 
-
 void YMenu::enableCommand(YAction action) {
-    for (int i = 0; i < itemCount(); i++)
-        if (action == actionNull || action == getItem(i)->getAction())
-            getItem(i)->setEnabled(true);
-    }
+    for (YMenuItem* item : fItems)
+        if (action == actionNull || action == item->getAction())
+            item->setEnabled(true);
+}
 
 void YMenu::disableCommand(YAction action) {
-    for (int i = 0; i < itemCount(); i++)
-        if (action == actionNull || action == getItem(i)->getAction())
-            getItem(i)->setEnabled(false);
-    }
+    for (YMenuItem* item : fItems)
+        if (action == actionNull || action == item->getAction())
+            item->setEnabled(false);
+}
 
 void YMenu::checkCommand(YAction action, bool check) {
-    for (int i = 0; i < itemCount(); i++)
-        if (action == actionNull || action == getItem(i)->getAction())
-            getItem(i)->setChecked(check);
+    for (YMenuItem* item : fItems)
+        if (action == actionNull || action == item->getAction())
+            item->setChecked(check);
+}
+
+void YMenu::removeCommand(YAction action) {
+    int i = -1;
+    for (YMenuItem* item : fItems)
+        if (++i, action == item->getAction())
+            return fItems.remove(i);
 }
 
 void YMenu::getOffsets(int &left, int &top, int &right, int &bottom) {
@@ -854,16 +907,16 @@ void YMenu::sizePopup(int hspace) {
 
     int height = t;
 
-    for (int i = 0; i < itemCount(); i++) {
-        const YMenuItem *mitem = getItem(i);
-
+    int count = 0;
+    for (YMenuItem* mitem : fItems) {
         int top, bottom, pad;
         int ih = mitem->queryHeight(top, bottom, pad);
 
+        ++count;
         if (height + ih >= 16000) {
             // evade library bug
-            TLOG(("truncating menu to %d items at height %d", i, height));
-            fItems.shrink(i);
+            TLOG(("truncating menu to %d items at height %d", count, height));
+            fItems.shrink(count);
             break;
         }
 
@@ -907,6 +960,12 @@ void YMenu::sizePopup(int hspace) {
     setSize(unsigned(width), unsigned(height));
 }
 
+void YMenu::repaintItem(YMenuItem* item) {
+    int id = find(fItems, item);
+    if (0 <= id)
+        repaintItem(id);
+}
+
 void YMenu::repaintItem(int item) {
     int x, y;
     unsigned h;
@@ -992,196 +1051,171 @@ void YMenu::drawSeparator(Graphics &g, int x, int y, unsigned w) {
     }
 }
 
-void YMenu::paintItem(Graphics &g, const int i, const int l, const int t, const int r,
-                      const int minY, const int maxY, bool draw) {
-    int const fontHeight(menuFont->height() + 1);
-    int const fontBaseLine(menuFont->ascent());
+void YMenu::paintItem(Graphics &g, const int i, const int l, const int t,
+                      const int r, const int minY, const int maxY, const int eh,
+                      const int top, const int bottom, const int pad)
+{
+    int const fontHeight(menuFont ? menuFont->height() + 1 : 1);
+    int const fontBaseLine(menuFont ? menuFont->ascent() : 1);
 
     YMenuItem *mitem = getItem(i);
-    mstring name = mitem->getName();
-    mstring param = mitem->getParam();
+    mstring& name = mitem->getName();
+    mstring& param = mitem->getParam();
 
-    if (mitem->getName() == null && mitem->getSubmenu() == nullptr) {
-        if (draw && t + 4 >= minY && t <= maxY)
-            drawSeparator(g, 1, t, width() - 2);
-    } else {
-        bool const active(i == paintedItem &&
-                          (mitem->getAction() != actionNull ||
-                           mitem->getSubmenu()));
+    const bool active(i == paintedItem &&
+                      (mitem->getAction() != actionNull || mitem->getSubmenu()));
+    const int ih = eh - top - bottom - pad - pad;
+    const int cascadePos(width() - r - 2 - ih - pad);
 
-        int eh, top, bottom, pad, ih;
-        eh = getItem(i)->queryHeight(top, bottom, pad);
-        ih = eh - top - bottom - pad - pad;
-
-        if (t + eh >= minY && t <= maxY) {
-            int const cascadePos(width() - r - 2 - ih - pad);
-            g.setColor((active && activeMenuItemBg) ? activeMenuItemBg : menuBg);
-
-            if (draw) {
-                if (active) {
-                    if (menuselPixbuf != null) {
-                        int ew = width() - r - l;
-                        if (fMenusel == null ||
-                            fMenusel->width() != unsigned(ew) ||
-                            fMenusel->height() != unsigned(eh))
-                        {
-                            ref<YImage> im = menuselPixbuf->scale(ew, eh);
-                            if (im != null) {
-                                fMenusel = im->renderToPixmap(depth());
-                            }
-                        }
-                        g.copyPixmap(fMenusel, 0, 0, ew, eh, l, t);
-                    }
-                    else if (menuselPixmap != null) {
-                        g.fillPixmap(menuselPixmap, l, t, width() - r - l, eh);
-                    } else if (activeMenuItemBg) {
-                        g.setColor(activeMenuItemBg);
-                        g.fillRect(l, t, width() - r - l, eh);
-                    } else {
-                        g.setColor(menuBg);
-                        drawBackground(g, l, t, width() - r - l, eh);
-                    }
-                } else {
-                    g.setColor(menuBg);
-                    drawBackground(g, l, t, width() - r - l, eh);
-                }
-
-                if ((wmLook == lookMetal || wmLook == lookFlat) && i != selectedItem) {
-                    g.setColor(menuBg->brighter());
-                    g.drawLine(1, t, 1, t + eh - 1);
-                    g.setColor(menuBg);
-                }
-
-                if (wmLook != lookWin95 && wmLook != lookWarp4 && active) {
-#ifdef OLDMOTIF
-                    bool raised(wmLook == lookMotif);
-#else
-                    bool raised(false);
-#endif
-                    g.setColor(menuBg);
-                    if (wmLook == lookGtk) {
-                        g.drawBorderW(l, t, width() - r - l - 1, eh - 1, true);
-                    } else if (wmLook == lookMetal || wmLook == lookFlat) {
-                        g.setColor((activeMenuItemBg ? activeMenuItemBg
-                                    : menuBg)->darker());
-                        g.drawLine(l, t, width() - r - l, t);
-                        g.setColor((activeMenuItemBg ? activeMenuItemBg
-                                    : menuBg)->brighter());
-                        g.drawLine(l, t + eh - 1, width() - r - l, t + eh - 1);
-                    } else
-                        g.draw3DRect(l, t, width() - r - l - 1, eh - 1, raised);
-
-
-                    if (wmLook == lookMotif)
-                        g.draw3DRect(l + 1, t + 1,
-                                     width() - r - l - 3, eh - 3, raised);
-                }
-
-                YColor fg(mitem->isEnabled() ? active ? activeMenuItemFg
-                           : menuItemFg
-                           : disabledMenuItemFg);
-                g.setColor(fg);
-                g.setFont(menuFont);
-
-                int delta = active;
-                if (wmLook == lookMotif || wmLook == lookGtk ||
-                    wmLook == lookWarp4 || wmLook == lookWin95 ||
-                    wmLook == lookMetal || wmLook == lookFlat)
-                    delta = 0;
-                int baseLine = t + top + pad + (ih - fontHeight) / 2 + fontBaseLine + delta;
-
-                if (mitem->isChecked()) {
-                    XPoint points[4];
-
-                    points[0].x = l + 3 + (16 - 10) / 2;
-                    points[1].x = 5;
-                    points[2].x = 5;
-                    points[3].x = -5;
-                    points[0].y = t + top + pad + ih / 2;
-                    points[1].y = -5;
-                    points[2].y = 5;
-                    points[3].y = 5;
-
-                    g.fillPolygon(points, 4, Convex, CoordModePrevious);
-                } else if (mitem->getIcon() != null) {
-                    int size = YIcon::menuSize();
-                    int dx = l + 1 + delta;
-                    int dy = t + delta + top + pad +
-                               (eh - top - pad * 2 - bottom - size) / 2;
-                    mitem->getIcon()->draw(g, dx, dy, size);
-                }
-
-                if (name != null) {
-                    int const maxWidth =
-                        (param != null ? paramPos - delta
-                         : mitem->getSubmenu() ? cascadePos : width()) -
-                        namePos;
-
-                    if (!mitem->isEnabled()) {
-                        g.setColor(disabledMenuItemSt ? disabledMenuItemSt :
-                                   menuBg->brighter());
-                        g.drawStringEllipsis(1 + delta + namePos, 1 + baseLine,
-                                             name, maxWidth);
-
-                        if (mitem->getHotCharPos() != -1)
-                            g.drawCharUnderline(1 + delta +  namePos, 1 + baseLine,
-                                                name, mitem->getHotCharPos());
-                    }
-                    g.setColor(fg);
-                    g.drawStringEllipsis(delta + namePos, baseLine, name, maxWidth);
-
-                    if (mitem->getHotCharPos() != -1)
-                        g.drawCharUnderline(delta + namePos, baseLine,
-                                            name, mitem->getHotCharPos());
-                }
-
-                if (param != null) {
-                    if (!mitem->isEnabled()) {
-                        g.setColor(disabledMenuItemSt ? disabledMenuItemSt :
-                                   menuBg->brighter());
-                        g.drawChars(param,
-                                    paramPos + delta + 1,
-                                    baseLine + 1);
-                    }
-                    g.setColor(fg);
-                    g.drawChars(param,
-                                paramPos + delta, baseLine);
-                } else if (mitem->getSubmenu() != nullptr) {
-                    if (mitem->getAction() != actionNull) {
-                        g.setColor(menuBg);
-                        if (false) {
-                            drawBackground(g, width() - r - 1 -ih - pad, t + top + pad, ih, ih);
-                            g.drawBorderW(width() - r - 1 - ih - pad, t + top + pad, ih - 1, ih - 1,
-                                          active ? false : true);
-                        } else {
-                            g.setColor(menuBg->darker());
-                            g.drawLine(cascadePos, t + top + pad,
-                                       cascadePos, t + top + pad + ih);
-                            g.setColor(menuBg->brighter());
-                            g.drawLine(cascadePos + 1, t + top + pad,
-                                       cascadePos + 1, t + top + pad + ih);
-
-                        }
-                        delta = (delta && active);
-                    }
-
-                    if (wmLook == lookGtk || wmLook == lookMotif) {
-                        int asize = 9;
-                        int ax = delta + width() - r - 1 - asize * 3 / 2;
-                        int ay = delta + t + top + pad + (ih - asize) / 2;
-                        g.setColor(menuBg);
-                        g.drawArrow(Right, ax, ay, asize, active);
-                    } else {
-                        int asize = 9;
-                        int ax = width() - r - 1 - asize;
-                        int ay = delta + t + top + pad + (ih - asize) / 2;
-
-                        g.setColor(fg);
-
-                        g.drawArrow(Right, ax, ay, asize);
-                    }
+    g.setColor((active && activeMenuItemBg) ? activeMenuItemBg : menuBg);
+    if (active) {
+        if (menuselPixbuf != null) {
+            int ew = width() - r - l;
+            if (fMenusel == null ||
+                fMenusel->width() != unsigned(ew) ||
+                fMenusel->height() != unsigned(eh))
+            {
+                ref<YImage> im = menuselPixbuf->scale(ew, eh);
+                if (im != null) {
+                    fMenusel = im->renderToPixmap(depth());
                 }
             }
+            g.copyPixmap(fMenusel, 0, 0, ew, eh, l, t);
+        }
+        else if (menuselPixmap != null) {
+            g.fillPixmap(menuselPixmap, l, t, width() - r - l, eh);
+        }
+        else if (activeMenuItemBg) {
+            g.setColor(activeMenuItemBg);
+            g.fillRect(l, t, width() - r - l, eh);
+        } else {
+            g.setColor(menuBg);
+            drawBackground(g, l, t, width() - r - l, eh);
+        }
+    } else {
+        g.setColor(menuBg);
+        drawBackground(g, l, t, width() - r - l, eh);
+    }
+
+    if (LOOK(lookMetal | lookFlat) && i != selectedItem) {
+        g.setColor(menuBg->brighter());
+        g.drawLine(1, t, 1, t + eh - 1);
+        g.setColor(menuBg);
+    }
+
+    if (wmLook != lookWin95 && wmLook != lookWarp4 && active) {
+#ifdef OLDMOTIF
+        bool raised(wmLook == lookMotif);
+#else
+        bool raised(false);
+#endif
+        g.setColor(menuBg);
+        if (wmLook == lookGtk) {
+            g.drawBorderW(l, t, width() - r - l - 1, eh - 1, true);
+        }
+        else if (wmLook == lookMetal || wmLook == lookFlat) {
+            g.setColor((activeMenuItemBg ? activeMenuItemBg
+                        : menuBg)->darker());
+            g.drawLine(l, t, width() - r - l, t);
+            g.setColor((activeMenuItemBg ? activeMenuItemBg
+                        : menuBg)->brighter());
+            g.drawLine(l, t + eh - 1, width() - r - l, t + eh - 1);
+        } else {
+            g.draw3DRect(l, t, width() - r - l - 1, eh - 1, raised);
+        }
+
+        if (wmLook == lookMotif)
+            g.draw3DRect(l + 1, t + 1,
+                         width() - r - l - 3, eh - 3, raised);
+    }
+
+    YColor fg(mitem->isEnabled() ? active ? activeMenuItemFg
+               : menuItemFg
+               : disabledMenuItemFg);
+    g.setColor(fg);
+    if (menuFont) {
+        g.setFont(menuFont);
+    }
+
+    int delta = active;
+    if (LOOK(lookMotif|lookGtk|lookWarp4|lookWin95|lookMetal|lookFlat))
+        delta = 0;
+    int baseLine = t + top + pad + (ih - fontHeight) / 2 + fontBaseLine + delta;
+    if (mitem->isChecked()) {
+        XPoint points[4] = {
+            { short(l + 3 + (16 - 10) / 2),
+              short(t + top + pad + ih / 2) },
+            { 5, -5 }, { 5, 5 }, { -5, 5 } };
+        g.fillPolygon(points, 4, Convex, CoordModePrevious);
+    }
+    else if (mitem->getIcon() != null) {
+        int size = YIcon::menuSize();
+        int dx = l + 1 + delta;
+        int dy = t + delta + top + pad +
+                   (eh - top - pad * 2 - bottom - size) / 2;
+        mitem->getIcon()->draw(g, dx, dy, size);
+    }
+
+    if (name != null) {
+        const int maxWidth = min(paramPos - namePos - 8,
+            (param != null ? paramPos - delta
+             : mitem->getSubmenu() ? cascadePos : int(width())) -
+            namePos);
+
+        if (!mitem->isEnabled()) {
+            g.setColor(disabledMenuItemSt ? disabledMenuItemSt :
+                       menuBg->brighter());
+            g.drawStringEllipsis(1 + delta + namePos, 1 + baseLine,
+                                 name, maxWidth);
+            if (mitem->getHotCharPos() != -1)
+                g.drawCharUnderline(1 + delta + namePos, 1 + baseLine,
+                                    name, mitem->getHotCharPos());
+        }
+        g.setColor(fg);
+        g.drawStringEllipsis(delta + namePos, baseLine, name, maxWidth);
+        if (mitem->getHotCharPos() != -1)
+            g.drawCharUnderline(delta + namePos, baseLine,
+                                name, mitem->getHotCharPos());
+    }
+
+    if (param != null) {
+        if (!mitem->isEnabled()) {
+            g.setColor(disabledMenuItemSt ? disabledMenuItemSt :
+                       menuBg->brighter());
+            g.drawChars(param, paramPos + delta + 1, baseLine + 1);
+        }
+        g.setColor(fg);
+        g.drawChars(param, paramPos + delta, baseLine);
+    }
+    else if (mitem->getSubmenu() != nullptr) {
+        if (mitem->getAction() != actionNull) {
+            g.setColor(menuBg);
+            if (false) {
+                drawBackground(g, width() - r - 1 -ih - pad, t + top + pad, ih, ih);
+                g.drawBorderW(width() - r - 1 - ih - pad, t + top + pad, ih - 1, ih - 1,
+                              active ? false : true);
+            } else {
+                g.setColor(menuBg->darker());
+                g.drawLine(cascadePos, t + top + pad,
+                           cascadePos, t + top + pad + ih);
+                g.setColor(menuBg->brighter());
+                g.drawLine(cascadePos + 1, t + top + pad,
+                           cascadePos + 1, t + top + pad + ih);
+            }
+            delta = (delta && active);
+        }
+        if (wmLook == lookGtk || wmLook == lookMotif) {
+            int asize = 9;
+            int ax = delta + width() - r - 1 - asize * 3 / 2;
+            int ay = delta + t + top + pad + (ih - asize) / 2;
+            g.setColor(menuBg);
+            g.drawArrow(Right, ax, ay, asize, active);
+        } else {
+            int asize = 9;
+            int ax = width() - r - 1 - asize;
+            int ay = delta + t + top + pad + (ih - asize) / 2;
+            g.setColor(fg);
+            g.drawArrow(Right, ax, ay, asize);
         }
     }
 }
@@ -1221,11 +1255,13 @@ void YMenu::paint(Graphics &g, const YRect &r1) {
 
     for (int i = 0; i < itemCount(); i++) {
         int ih = getItem(i)->queryHeight(top, bottom, pad);
-
-        if (iy < r1.y() + (int) r1.height() &&
-            iy + (int) ih > r1.y())
-        {
-            paintItem(g, i, l, iy, r, r1.y(), r1.y() + r1.height(), true);
+        if (iy < r1.y() + int(r1.height()) && iy + ih > r1.y()) {
+            if (ih <= 8) {
+                drawSeparator(g, 1, iy, width() - 2);
+            } else {
+                paintItem(g, i, l, iy, r, r1.y(), r1.y() + r1.height(),
+                          ih, top, bottom, pad);
+            }
         }
         iy += ih;
     }
